import math
from typing import Optional

import ee

"""
creating a tiler using earthengine classes, so it can be used on-cluster in a parallel way.
"""

TILE_SIZE: float = 256
ORIGIN: float = math.pi * 6378137  # earth depth
C: float = 40075016.686  # earth circumpherence
PROJECTION: str = "EPSG:3857"


def zoom_to_scale(zoom: ee.Number) -> ee.Number:
    zoom: ee.Number = ee.Number(zoom)
    return ee.Number(C).divide(ee.Number(2).pow(zoom)).divide(TILE_SIZE)


def scale_to_zoom(scale: ee.Number) -> ee.Number:
    scale: ee.Number = ee.Number(scale)
    zoom: ee.Number = (
        ee.Number(C).divide(scale.multiply(TILE_SIZE)).log().divide(ee.Number(2).log())
    )
    return zoom.ceil()


def to_radians(degrees: ee.Number) -> ee.Number:
    degrees: ee.Number = ee.Number(degrees)

    return degrees.multiply(math.pi).divide(180)


def pixels_to_meters(px: ee.Number, py: ee.Number, zoom: ee.Number) -> ee.List:
    px: ee.Number = ee.Number(px)
    py: ee.Number = ee.Number(py)
    zoom = ee.Number(zoom)

    resolution: ee.Number = zoom_to_scale(zoom)
    x: ee.Number = px.multiply(resolution).subtract(ORIGIN)
    y: ee.Number = py.multiply(resolution).subtract(ORIGIN)
    return ee.List([x, y])


def meters_to_pixels(x: ee.Number, y: ee.Number, zoom: ee.Number) -> ee.List:
    x: ee.Number = ee.Number(x)
    y: ee.Number = ee.Number(y)
    zoom = ee.Number(zoom)

    resolution: ee.Number = zoom_to_scale(zoom)
    px: ee.Number = x.add(ORIGIN).divide(resolution)
    py: ee.Number = y.add(ORIGIN).divide(resolution)
    return ee.List([px, py])


def degrees_to_tiles(lon: ee.Number, lat: ee.Number, zoom: ee.Number) -> ee.List:
    lon: ee.Number = ee.Number(lon)
    lat: ee.Number = ee.Number(lat)
    zoom: ee.Number = ee.Number(zoom)

    tx: ee.Number = lon.add(180).divide(360).multiply(ee.Number(2).pow(zoom)).floor()
    ty: ee.Number = (
        ee.Number(1)
        .subtract(
            to_radians(lat)
            .tan()
            .add(ee.Number(1).divide(to_radians(lat).cos()))
            .log()
            .divide(ee.Number(math.pi))
        )
        .divide(2)
        .multiply(ee.Number(2).pow(zoom))
        .floor()
    )
    return ee.List([tx, ty])


def get_tile_bounds(tx: ee.Number, ty: ee.Number, zoom: ee.Number) -> ee.List:
    tx: ee.Number = ee.Number(tx)
    ty: ee.Number = ee.Number(ty)
    zoom: ee.Number = ee.Number(zoom)

    ty_flip: ee.Number = (
        ee.Number(2).pow(zoom).subtract(ty).subtract(1)
    )  # TMS -> XYZ, flip y index
    min: ee.Number = pixels_to_meters(
        ee.Number(tx).multiply(TILE_SIZE), ty_flip.multiply(TILE_SIZE), zoom
    )
    max: ee.Number = pixels_to_meters(
        ee.Number(tx).add(1).multiply(TILE_SIZE),
        ty_flip.add(1).multiply(TILE_SIZE),
        zoom,
    )
    return ee.List([min, max])


def get_tiles_for_geometry(
    geometry: ee.Geometry, zoom: ee.Number, opt_bounds: Optional[ee.Geometry] = None
) -> ee.FeatureCollection:
    zoom: ee.Number = ee.Number(zoom)

    bounds: ee.Geometry = ee.Geometry(
        ee.Algorithms.If(opt_bounds, opt_bounds, geometry)
    )
    bounds_list: ee.List = ee.List(bounds.bounds().coordinates().get(0))

    ll: ee.List = ee.List(bounds_list.get(0))
    ur: ee.List = ee.List(bounds_list.get(2))

    tmin: ee.List = ee.List(
        degrees_to_tiles(ee.Number(ll.get(0)), ee.Number(ll.get(1)), zoom)
    )
    tmax: ee.List = ee.List(
        degrees_to_tiles(ee.Number(ur.get(0)), ee.Number(ur.get(1)), zoom)
    )

    # create ranges for server-side mapping
    tx_range: ee.List = ee.List.sequence(tmin.get(0), ee.Number(tmax.get(0)).add(1))
    ty_range: ee.List = ee.List.sequence(tmax.get(1), ee.Number(tmin.get(1)).add(1))

    def create_tile(tx: ee.Number, ty: ee.Number, zoom: ee.Number) -> ee.Feature:
        """
        Add a tile to input tiles
        """
        tx: ee.Number = ee.Number(tx)
        ty: ee.Number = ee.Number(ty)
        zoom: ee.Number = ee.Number(zoom)

        tile_bounds: ee.List = get_tile_bounds(tx, ty, zoom)
        rect: ee.Geometry = ee.Geometry.Rectangle(tile_bounds, PROJECTION, False)
        return ee.Feature(rect).set(
            {"tx": tx.format(), "ty": ty.format(), "zoom": zoom.format()}
        )

    tiles: ee.List = ee.List(
        tx_range.map(
            lambda tx: ty_range.map(lambda ty: create_tile(tx=tx, ty=ty, zoom=zoom))
        )
    ).flatten()

    return ee.FeatureCollection(tiles).filterBounds(geometry)



import logging
from pathlib import Path
import requests
import shutil
from typing import Any, Callable, Dict, Optional

from pathos import logger
from pathos.core import getpid
from retry import retry


@retry(tries=10, delay=5, backoff=10)
def download_image(
    i_ee,
    index: int,
    image_download_method: Callable,
    serialized_image_list: Dict[str, Any],
    name_prefix: str,
    out_dir: Optional[Path],
    download_kwargs: Optional[Dict[str, Any]],
) -> None:
    """
    Hidden function to be used with download_image_collection. As we want compatibility with
    windows, we need to use dill to pickle the initialized earthengine module. Pathos uses
    dill pickle data uses to start new processes with its multiprocessing pool, while the python
    multiprocessing module uses pickle, which cannot pickle module objects.
    """

    if not out_dir:
        out_dir: Path = Path.cwd() / "output"
    if not download_kwargs:
        download_kwargs: Dict[str, str] = {}

    pid = getpid()
    log_file: Path = out_dir / "logging" / f"{pid}.log"
    fh = logging.FileHandler(log_file)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
    fh.setFormatter(formatter)

    plogger = logger(level=logging.DEBUG, handler=fh)

    ee_api_url: str = i_ee.data._cloud_api_base_url
    ee_hv_url: str = "https://earthengine-highvolume.googleapis.com"
    if ee_api_url is not ee_hv_url:
        i_ee.Initialize(opt_url=ee_hv_url)

    image_list: i_ee.List = i_ee.deserializer.fromJSON(serialized_image_list)

    img: i_ee.Image = i_ee.Image(image_list.get(index))
    url: str = image_download_method(img, download_kwargs)
    r: requests.Response = requests.get(url, stream=True)

    # File format chain
    format: Optional[str] = download_kwargs.get("format")
    if format:
        if format == "GEO_TIFF":
            extention: str = ".tif"
        elif format == "NPY":
            extention: str = ".npy"
        elif format == "PNG":
            extention: str = ".png"
        elif format == "JPG":
            extention: str = ".jpg"
        else:
            extention: str = ".tif.zip"
    elif image_download_method == i_ee.Image.getDownloadURL:
        extention: str = ".tif.zip"
    elif image_download_method == i_ee.Image.getThumbURL:
        extention: str = ".png"
    else:
        raise RuntimeError(f"image download method {image_download_method} unknown.")

    # Image naming chain
    img_props: Dict[str, Any] = img.getInfo()["properties"]
    t0: Optional[int] = img_props.get("system:time_start")
    img_index: Optional[int] = img_props.get("system:index")
    if t0:
        file_id: str = t0
    elif img_index:
        file_id: str = img_index
    else:
        file_id: str = index

    # File name
    filename: Path = out_dir / f"{name_prefix}{file_id}{extention}"

    with open(filename, "wb") as out_file:
        shutil.copyfileobj(r.raw, out_file)

    plogger.info(f"Done: {index}")
    plogger.removeHandler(fh)



from typing import Any, Dict, List, Optional
from datetime import datetime

import ee

from eepackages import assets
from eepackages import gl
from eepackages import utils


class Bathymetry(object):
    def __init__(self, waterIndexMin: float = -0.15, waterIndexMax: float = 0.35):
        self.waterIndexMin = waterIndexMin
        self.waterIndexMax = waterIndexMax

    @staticmethod
    def _remove_all_zero_images(image: ee.Image):
        mask: ee.Image = (
            image.select(["blue", "green", "red", "nir", "swir"])
            .mask()
            .reduce(ee.Reducer.allNonZero())
        )
        return image.updateMask(mask)

    def compute_inverse_depth(
        self,
        bounds,
        start: datetime,
        stop: datetime,
        filter_masked: bool,
        scale: float,
        missions: List[str] = ["S2", "L8"],
        cloud_frequency_threshold_data: float = 0.15,
        pansharpen: bool = False,
        skip_neighborhood_search: bool = False,
        skip_scene_boundary_fix: bool = False,
        bounds_buffer: int = 10000,
    ) -> ee.Image:
        images: ee.ImageCollection = self.get_images(
            bounds=bounds,
            start=start,
            stop=stop,
            filter_masked=filter_masked,
            scale=scale,
            missions=missions,
            cloud_frequency_threshold_delta=cloud_frequency_threshold_data,
        )
        # save loaded images in class as raw_images
        self._raw_images = images
        images = images.map(self._remove_all_zero_images)

        bounds = bounds.buffer(bounds_buffer, ee.Number(bounds_buffer).divide(10))

        return self._compute_inverse_depth(
            images=images,
            bounds=bounds,
            scale=scale,
            pansharpen=pansharpen,
            skip_neighborhood_search=skip_neighborhood_search,
            skip_scene_boundary_fix=skip_scene_boundary_fix,
        )

    def _compute_inverse_depth(
        self,
        images: ee.ImageCollection,
        bounds,
        scale: int,
        pansharpen: bool,
        skip_neighborhood_search: bool,
        skip_scene_boundary_fix: bool,
    ) -> ee.Image:
        bands: List[str] = ["red", "green", "blue"]
        green_max: float = 0.4

        def _set_image_area_properties(image: ee.Image) -> ee.Image:
            water: ee.Image = (
                image.normalizedDifference(["green", "nir"])
                .rename("water")
                .unitScale(0, 0.1)
            )

            if pansharpen:
                # TODO: Not implemented
                raise NotImplementedError()
                image = assets.pansharpen(image)

            water_area: ee.Number = (
                water.gt(0.01)
                .multiply(ee.Image.pixelArea())
                .reduceRegion(
                    reducer=ee.Reducer.sum(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(5),
                    tileScale=4,
                )
                .values()
                .get(0)
            )

            land_area: ee.Number = (
                water.lt(0)
                .multiply(ee.Image.pixelArea())
                .reduceRegion(
                    reducer=ee.Reducer.sum(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(5),
                    tileScale=4,
                )
                .values()
                .get(0)
            )

            dark: ee.Image = image

            dark = dark.updateMask(water.gt(0)).reduceRegion(
                reducer=ee.Reducer.percentile([0]),
                geometry=bounds,
                scale=scale,
                maxPixels=1e10,
                tileScale=4,
            )

            image = image.set(dark).set(
                {"water": water, "waterArea": water_area, "landArea": land_area}
            )

            return image

        images: ee.ImageCollection = images.map(_set_image_area_properties)
        self._images_area_properties = images

        # Filter images with negative RGB values
        images = images.filter(
            ee.Filter.And(
                ee.Filter.gt(bands[0], 0),
                ee.Filter.gt(bands[1], 0),
                ee.Filter.gt(bands[2], 0),
            )
        )

        if skip_neighborhood_search:
            images = assets.addCdfQualityScore(
                images=images,
                opt_thresholdMin=70,
                opt_thresholdMax=80,
                opt_includeNeighborhood=False,
            )
        else:
            images = assets.addCdfQualityScore(
                images=images,
                opt_thresholdMin=70,
                opt_thresholdMax=80,
                opt_includeNeighborhood=True,
                opt_neighborhoodOptions={"erosion": 0, "dilation": 0, "weight": 200},
            )

        def fix_scene_boundaries(image: ee.Image) -> ee.Image:
            weight: ee.Image = image.select("weight")
            mask = image.select(0).mask()

            mask: ee.Image = (
                utils.focalMin(mask, 10)
                .reproject(ee.Projection("EPSG:3857").atScale(scale))
                .resample("bicubic")
            )
            mask = (
                utils.focalMaxWeight(mask.Not(), 10)
                .reproject(ee.Projection("EPSG:3857").atScale(scale))
                .resample("bicubic")
            )
            mask = ee.Image.constant(1).subtract(mask)

            weight = weight.multiply(mask)

            return image.addBands(weight, None, True)

        if not skip_scene_boundary_fix:
            images = images.map(fix_scene_boundaries)

        def image_map_func(i: ee.Image):
            t: str = i.get("system:time_start")
            weight: ee.Image = i.select("weight")

            dark_image: ee.Image = ee.Image.constant(
                list(map(lambda n: i.get(n), bands))
            ).rename(bands)
            mission: str = i.get("MISSION")
            scale_water_to: str = "percentiles"
            scale_land_to: str = "percentiles"

            range_percentiles_water: List[int] = [2, 98]
            range_percentiles_land: List[int] = [2, 98]

            range_sigma_water: List[int] = [1, 1]
            range_sigma_land: List[int] = [2, 2]

            water: ee.Image = ee.Image(i.get("water"))
            non_water: ee.Image = water.subtract(1).multiply(-1)  # Not being used

            i = i.select(bands).subtract(dark_image).max(0.0001)

            i_all: ee.Image = i

            water2: ee.Image = gl.smoothStep(-0.05, 0.2, water)
            non_water2: ee.Image = water2.subtract(1).multiply(-1)

            i = i.log()

            stat1: ee.Image = i

            stat1 = stat1.updateMask(
                water2.multiply(i_all.select("green").lt(green_max))
            )

            if scale_water_to == "percentiles":
                stat1 = stat1.reduceRegion(
                    reducer=ee.Reducer.percentile(range_percentiles_water),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )

                min1: List[str] = [
                    ee.String(stat1.get(f"{band}_p{range_percentiles_water[0]}"))
                    for band in bands
                ]
                max1: List[str] = [
                    ee.String(stat1.get(f"{band}_p{range_percentiles_water[1]}"))
                    for band in bands
                ]

            if scale_water_to == "sigma":
                stat1mean = stat1.reduceRegion(
                    reducer=ee.Reducer.mean(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )

                stat1sigma = stat1.reduceregion(  # not being used.
                    reducer=ee.Reducer.stDev(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )

                # Not sure whether this si tested, min1 should always be zero
                min1: List[ee.Number] = [
                    ee.Number(stat1mean.get(band)).subtract(
                        ee.Number(stat1mean.get(band)).multiply(range_sigma_water[0])
                    )
                    for band in bands
                ]
                max1: List[ee.Number] = [
                    ee.Number(stat1mean.get(band)).add(
                        ee.Number(stat1mean.get(band)).multiply(range_sigma_water[1])
                    )
                    for band in bands
                ]

            min1 = self._fix_null(min1, 0)
            max1 = self._fix_null(max1, 0.001)

            stat2: ee.Image = i_all.updateMask(
                non_water2.multiply(i_all.select("green").lt(green_max))
            )

            if scale_land_to == "percentiles":
                stat2 = stat2.reduceRegion(
                    reducer=ee.Reducer.percentile(range_percentiles_land),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )

                min2 = [
                    ee.String(stat2.get(f"{band}_p{range_percentiles_land[0]}"))
                    for band in bands
                ]
                max2 = [
                    ee.String(stat2.get(f"{band}_p{range_percentiles_land[1]}"))
                    for band in bands
                ]

            if scale_land_to == "sigma":
                stat2mean = stat2.reduceRegion(
                    reducer=ee.Reducer.mean(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )
                stat2sigma = stat2.reduceRegion(  # not used?
                    reducer=ee.Reducer.stDev(),
                    geometry=bounds,
                    scale=ee.Number(scale).multiply(3),
                    maxPixels=1e10,
                )

                min2: List[ee.Number] = [
                    ee.Number(stat2mean)
                    .get(band)
                    .subtract(
                        ee.Number(stat2mean.get(band)).multiply(range_sigma_land[0])
                    )
                    for band in bands
                ]
                max2: List[ee.Number] = [
                    ee.Number(stat2mean.get(band)).add(
                        ee.Number(stat2mean.get(band)).multiply(range_sigma_land[1])
                    )
                    for band in bands
                ]

            min2 = self._fix_null(min2, 0)
            max2 = self._fix_null(max2, 0.001)

            i_water = self._unit_scale(i.select(bands), min1, max1).updateMask(water2)

            i_land = self._unit_scale(i_all.select(bands), min2, max2).updateMask(
                non_water2
            )

            i = i_water.blend(i_land).addBands(water)

            i = i.addBands(weight)

            return i.set(
                {
                    "label": ee.Date(t).format().cat(", ").cat(mission),
                    "system:time_start": t,
                }
            )

        images = images.map(image_map_func)

        self._images_with_statistics = images

        # mean = sum(w * x) / sum(w)
        image: ee.Image = (
            images.map(
                lambda i: i.select(bands + ["water"]).multiply(i.select("weight"))
            )
            .sum()
            .divide(images.select("weight").sum())
        )

        return image

    def compute_intertidal_depth(
        self,
        bounds,
        start: datetime,
        stop: datetime,
        scale: float,
        filter_masked: bool,
        filter_masked_fraction: Optional[float] = None,
        filter: Optional[ee.Filter] = None,
        bounds_buffer: Optional[float] = None,
        water_index_min: Optional[float] = None,
        water_index_max: Optional[float] = None,
        missions: List[str] = ["S2", "L8"],
        skip_scene_boundary_fix=False,
        skip_neighborhood_search: bool = False,
        neighborhood_search_parameters: Dict[str, float] = {
            "erosion": 0,
            "dilation": 0,
            "weight": 100,
        },
        lower_cdf_boundary: float = 70,
        upper_cdf_boundary: float = 80,
        cloud_frequency_threshold_data: float = 0.15,
        clip: bool = False,
    ) -> ee.Image:
        if water_index_min:
            self.waterIndexMin = water_index_min
        water_index_min = self.waterIndexMin
        if water_index_max:
            self.waterIndexMax = water_index_max
        water_index_max = self.waterIndexMax
        if bounds_buffer:
            bounds = bounds.buffer(bounds_buffer, bounds_buffer / 10)

        images: ee.ImageCollection = self.get_images(
            bounds=bounds,
            start=start,
            stop=stop,
            filter_masked=filter_masked,
            filter_masked_fraction=filter_masked_fraction,
            scale=scale,
            missions=missions,
            cloud_frequency_threshold_delta=cloud_frequency_threshold_data,
            filter=filter,
        )

        self._raw_images = images

        bands: List[str] = ["blue", "green", "red", "nir", "swir"]

        # Mask all zero images
        def mask_zero_images(i: ee.Image) -> ee.Image:
            mask: ee.Image = (
                i.select(bands).mask().reduce(ee.Reducer.allNonZero()).eq(1)
            )
            return i.updateMask(mask)

        images = images.map(mask_zero_images)

        if not bounds_buffer:
            bounds_buffer = 10000  # Not being used.

        if skip_neighborhood_search:
            images = assets.addCdfQualityScore(images, 70, 80, False)
        else:
            images = assets.addCdfQualityScore(
                images=images,
                opt_thresholdMin=lower_cdf_boundary,
                opt_thresholdMax=upper_cdf_boundary,
                opt_includeNeighborhood=True,
                opt_neighborhoodOptions=neighborhood_search_parameters,
            )

        if not skip_scene_boundary_fix:

            def fix_scene_boundaries(i: ee.Image) -> ee.Image:
                weight: ee.Image = i.select("weight")
                mask: ee.Image = i.select(0).mask()

                mask = (
                    utils.focalMin(mask, 10)
                    .reproject(ee.Projection("EPSG:3857").atScale(scale))
                    .resample("bicubic")
                )
                mask = (
                    utils.focalMaxWeight(mask.Not(), 10)
                    .reproject(ee.Projection("EPSG:3857").atScale(scale))
                    .resample("bicubic")
                )
                mask = ee.Image.constant(1).subtract(mask)

                weight = weight.multiply(mask)

                return i.addBands(srcImg=weight, overwrite=True)

            images = images.map(fix_scene_boundaries)

        self._refined_images = images

        # mean = sum(w * x) / sum (w)
        self.composite = (
            images.map(lambda i: i.select(bands).multiply(i.select("weight")))
            .sum()
            .divide(images.select("weight").sum())
            .select(["red", "green", "blue", "swir", "nir"])
        )

        bands = ["ndwi", "indvi", "mndwi"]

        def calculate_water_statistics(i: ee.Image) -> ee.Image:
            t = i.get("system:time_start")  # Not used
            weight: ee.Image = i.select("weight")

            ndwi: ee.Image = i.normalizedDifference(["green", "nir"]).rename("ndwi")
            indvi: ee.Image = i.normalizedDifference(["red", "nir"]).rename("indvi")
            mndwi: ee.Image = i.normalizedDifference(["green", "swir"]).rename("mndwi")

            ndwi = ndwi.clamp(water_index_min, water_index_max)
            indvi = indvi.clamp(water_index_min, water_index_max)
            mndwi = mndwi.clamp(water_index_min, water_index_max)

            return ee.Image([ndwi, indvi, mndwi]).addBands(weight)

        images = images.map(calculate_water_statistics)

        self._images_with_statistics = images

        image = (
            images.map(lambda i: i.select(bands).multiply(i.select("weight")))
            .sum()
            .divide(images.select("weight").sum())
        )

        if clip == True:
            return image.clip(bounds)
        else:
            return image

    @staticmethod
    def get_images(
        bounds,
        start: datetime,
        stop: datetime,
        filter_masked: bool,
        scale: float,
        missions: List[str],
        filter_masked_fraction: Optional[float] = None,
        cloud_frequency_threshold_delta: float = 0.15,
        filter: Optional[ee.Filter] = None,
    ) -> ee.ImageCollection:
        date_filter: ee.Filter = ee.Filter.date(start, stop)
        if filter:
            filter: ee.Filter = ee.Filter.And(filter, date_filter)
        else:
            filter: ee.Filter = date_filter

        options_get_images: Dict[str, Any] = {
            "missions": missions,
            "filter": filter,
            "filterMasked": filter_masked,
            "filterMaskedFraction": filter_masked_fraction,
            "scale": ee.Number(scale).multiply(10),  # why *10?
            "resample": True,
        }

        images: ee.ImageCollection = assets.getImages(bounds, options_get_images)

        options_get_mostly_clean_images: Dict[str, Any] = {
            "cloudFrequencyThresholdDelta": cloud_frequency_threshold_delta
        }

        return assets.getMostlyCleanImages(
            images, bounds, options_get_mostly_clean_images
        )

    @staticmethod
    def _fix_null(values, v) -> ee.List:
        return ee.List(values).map(
            lambda o: ee.Algorithms.If(ee.Algorithms.IsEqual(o, None), v, o)
        )

    @staticmethod
    def _unit_scale(image: ee.Image, min: float, max: float) -> ee.Image:
        min_image: ee.Image = ee.Image.constant(min)
        max_image: ee.Image = ee.Image.constant(max)
        return image.subtract(min_image).divide(max_image.subtract(min_image))



import ee

from eepackages import utils


def cloudMaskAlgorithms_Landsat(image):
    imageWithCloud = ee.Algorithms.Landsat.simpleCloudScore(ee.Image(image))

    return imageWithCloud.addBands(
        ee.Image(imageWithCloud.select("cloud").divide(100)), None, True
    )


def cloudMaskAlgorithms_S2(image):
    qa = image.select("QA60")
    #    Bits 10 and 11 are clouds and cirrus, respectively.
    cloudBitMask = 1 << 10
    cirrusBitMask = 1 << 11
    # Both flags should be set to zero, indicating clear conditions.
    mask = qa.bitwiseAnd(cloudBitMask).eq(0).And(qa.bitwiseAnd(cirrusBitMask).eq(0))
    mask = mask.subtract(1).multiply(-1).rename("cloud")
    return image.addBands(mask)


cloudMaskAlgorithms = {
    "L8": cloudMaskAlgorithms_Landsat,
    "S2": cloudMaskAlgorithms_S2,
    "L7": cloudMaskAlgorithms_Landsat,
    "L5": cloudMaskAlgorithms_Landsat,
    "L4": cloudMaskAlgorithms_Landsat,
}


def getImages(g, options):
    g = ee.Geometry(g)

    resample = False

    s2MergeByTime = True

    if options and "s2MergeByTime" in options:
        s2MergeByTime = options["s2MergeByTime"]

    cloudMask = False

    if options and "cloudMask" in options:
        cloudMask = options["cloudMask"]

    cloudMaskAlgorithms_ = cloudMaskAlgorithms
    if options and "cloudMaskAlgorithms" in options:
        cloudMaskAlgorithms_ = options["cloudMaskAlgorithms"]

    missions = ["S2", "L8", "L9"]

    if options and "missions" in options:
        missions = options["missions"]

    if options and "resample" in options:
        resample = options["resample"]

    bands = {
        "S2": {
            "from": ["B11", "B8", "B4", "B3", "B2"],
            "to": ["swir", "nir", "red", "green", "blue"],
        },
        "L8": {
            "from": ["B6", "B5", "B4", "B3", "B2"],
            "to": ["swir", "nir", "red", "green", "blue"],
        },
        "L9": {
            "from": ["B6", "B5", "B4", "B3", "B2", "B6", "B7"],
            "to": ["swir", "nir", "red", "green", "blue", "swir1", "swir2"],
        },
        "L5": {
            "from": ["B5", "B4", "B3", "B2", "B1"],
            "to": ["swir", "nir", "red", "green", "blue"],
        },
        "L4": {
            "from": ["B5", "B4", "B3", "B2", "B1"],
            "to": ["swir", "nir", "red", "green", "blue"],
        },
        "L7": {
            "from": ["B5", "B4", "B3", "B2", "B1"],
            "to": ["swir", "nir", "red", "green", "blue"],
        },
    }

    if options and "includeTemperature" in options:
        bands["L8"]["from"].append("B10")
        bands["L8"]["to"].append("temp")
        bands["L9"]["from"].append("B10")
        bands["L9"]["to"].append("temp")
        bands["L5"]["from"].append("B6")
        bands["L5"]["to"].append("temp")
        bands["L4"]["from"].append("B6")
        bands["L4"]["to"].append("temp")
        bands["L7"]["from"].append("B6_VCID_1")
        bands["L7"]["to"].append("temp")

    # used only for a single sensor
    if options and "bandsAll" in options:
        bands["S2"] = {
            "from": [
                "B11",
                "B8",
                "B3",
                "B2",
                "B4",
                "B5",
                "B6",
                "B7",
                "B8A",
                "B9",
                "B10",
            ],
            "to": [
                "swir",
                "nir",
                "green",
                "blue",
                "red",
                "red2",
                "red3",
                "red4",
                "nir2",
                "water_vapour",
                "cirrus",
            ],
        }

    s2 = ee.ImageCollection("COPERNICUS/S2_HARMONIZED").filterBounds(g)

    if options and "filter" in options:
        s2 = s2.filter(options["filter"])

    # apply custom cloud masking
    if cloudMask:
        s2 = s2.map(cloudMaskAlgorithms_["S2"])
        bands["S2"]["from"].append("cloud")
        bands["S2"]["to"].append("cloud")

    s2 = s2.select(bands["S2"]["from"], bands["S2"]["to"])

    l9 = ee.ImageCollection("LANDSAT/LC09/C02/T1_TOA")

    l9 = l9.filterBounds(g)

    if options and "filter" in options:
        l9 = l9.filter(options["filter"])

    if cloudMask:
        l9 = l9.map(cloudMaskAlgorithms_["L9"])
        bands["L9"]["from"].append("cloud")
        bands["L9"]["to"].append("cloud")

    l9 = l9.select(bands["L9"]["from"], bands["L9"]["to"])

    l8 = ee.ImageCollection("LANDSAT/LC08/C01/T1_RT_TOA")

    if options and "includeTier2" in options:
        l8 = l8.merge(ee.ImageCollection("LANDSAT/LC08/C01/T2_TOA"))

    l8 = l8.filterBounds(g)

    if options and "filter" in options:
        l8 = l8.filter(options["filter"])

    if cloudMask:
        l8 = l8.map(cloudMaskAlgorithms_["L8"])
        bands["L8"]["from"].append("cloud")
        bands["L8"]["to"].append("cloud")

    l8 = l8.select(bands["L8"]["from"], bands["L8"]["to"])

    l5 = ee.ImageCollection("LANDSAT/LT05/C01/T1_TOA")

    if options and "includeTier2" in options:
        l5 = l5.merge(ee.ImageCollection("LANDSAT/LT05/C01/T2_TOA"))

    l5 = l5.filterBounds(g)

    if options and "filter" in options:
        l5 = l5.filter(options["filter"])

    if cloudMask:
        l5 = l5.map(cloudMaskAlgorithms_["L5"])
        bands["L5"]["from"].append("cloud")
        bands["L5"]["to"].append("cloud")

    l5 = l5.select(bands["L5"]["from"], bands["L5"]["to"])

    l4 = ee.ImageCollection("LANDSAT/LT04/C01/T1_TOA")

    if options and "includeTier2" in options:
        l4 = l4.merge(ee.ImageCollection("LANDSAT/LT04/C01/T2_TOA"))

    l4 = l4.filterBounds(g)

    if options and "filter" in options:
        l4 = l4.filter(options["filter"])

    if cloudMask:
        l4 = l4.map(cloudMaskAlgorithms_["L4"])
        bands["L4"]["from"].append("cloud")
        bands["L4"]["to"].append("cloud")

    l4 = l4.select(bands["L4"]["from"], bands["L4"]["to"])

    l7 = ee.ImageCollection("LANDSAT/LE07/C01/T1_RT_TOA")

    if options and "includeTier2" in options:
        l7 = l7.merge(ee.ImageCollection("LANDSAT/LE07/C01/T2_TOA"))

    l7 = l7.filterBounds(g)

    if options and "filter" in options:
        l7 = l7.filter(options["filter"])

    if cloudMask:
        l7 = l7.map(cloudMaskAlgorithms_["L7"])
        bands["L7"]["from"].append("cloud")
        bands["L7"]["to"].append("cloud")

    l7 = l7.select(bands["L7"]["from"], bands["L7"]["to"])

    if options and "clipBufferSizeL7" in options:

        def f(i):
            mask = (
                i.select(["green", "red", "nir", "swir"])
                .mask()
                .reduce(ee.Reducer.allNonZero())
            )
            mask = mask.focal_min(
                options["clipBufferSizeL7"], "square", "meters"
            ).reproject(i.select("nir").projection())
            return i.updateMask(mask)

        l7 = l7.map(f)

    clipBufferSize = 6000

    if options and "clipBufferSize" in options:
        clipBufferSize = options["clipBufferSize"]

    scale = 100
    if options and "scale" in options:
        scale = options["scale"]

    def clipNegativeFootprint(i):
        return i.clip(
            i.select(0).geometry().buffer(ee.Number(clipBufferSize).multiply(-1), 1000)
        )

    l4 = l4.map(clipNegativeFootprint)
    l5 = l5.map(clipNegativeFootprint)
    l7 = l7.map(clipNegativeFootprint)

    resample = "bicubic"

    s2 = s2.map(lambda i: i.resample(resample))

    # merge by time (remove duplicates)
    if s2MergeByTime:
        s2 = mosaicByTime(s2)

    def f2(i):
        return (
            i.addBands(i.multiply(0.0001).float(), i.bandNames().remove("cloud"), True)
            .copyProperties(i)
            .set(
                {
                    "SUN_ELEVATION": ee.Number(90).subtract(
                        i.get("MEAN_SOLAR_ZENITH_ANGLE")
                    )
                }
            )
            .set({"MISSION": "S2"})
            .set({"SUN_AZIMUTH": i.get("MEAN_SOLAR_AZIMUTH_ANGLE")})
            .set({"MULTIPLIER": 0.0001})
            .set({"SPACECRAFT_ID": i.get("SPACECRAFT_NAME")})
        )

    s2 = s2.map(f2)

    images = ee.ImageCollection([])

    if "L5" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "L5",
                    "BANDS_FROM": bands["L5"]["from"],
                    "BANDS_TO": bands["L5"]["to"],
                    "MULTIPLIER": 1,
                }
            )

        l5 = l5.map(f)
        images = images.merge(l5)

    if "L4" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "L4",
                    "BANDS_FROM": bands["L4"]["from"],
                    "BANDS_TO": bands["L4"]["to"],
                    "MULTIPLIER": 1,
                }
            )

        l4 = l4.map(f)
        images = images.merge(l4)

    if "L7" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "L7",
                    "BANDS_FROM": bands["L7"]["from"],
                    "BANDS_TO": bands["L7"]["to"],
                    "MULTIPLIER": 1,
                }
            )

        l7 = l7.map(f)
        images = images.merge(l7)

    if "L8" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "L8",
                    "BANDS_FROM": bands["L8"]["from"],
                    "BANDS_TO": bands["L8"]["to"],
                    "MULTIPLIER": 1,
                }
            )

        l8 = l8.map(f)
        images = images.merge(l8)

    if "L9" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "L9",
                    "BANDS_FROM": bands["L9"]["from"],
                    "BANDS_TO": bands["L9"]["to"],
                    "MULTIPLIER": 1,
                }
            )

        l9 = l9.map(f)
        images = images.merge(l9)

    images = ee.ImageCollection(images)

    if resample:
        images = images.map(lambda i: i.resample(resample))

    if "S2" in missions:

        def f(i):
            return i.set(
                {
                    "MISSION": "S2",
                    "BANDS_FROM": bands["S2"]["from"],
                    "BANDS_TO": bands["S2"]["to"],
                    "MULTIPLIER": 0.0001,
                }
            )

        s2 = s2.map(f)
        images = images.merge(s2)

    images = ee.ImageCollection(images)

    if options and "filterMasked" in options:
        if "filterMaskedFraction" in options and options["filterMaskedFraction"]:
            # get images coverting bounds at least filterMaskedFraction%
            area = g.area(scale)

            def f(i):
                maskedArea = (
                    ee.Image.pixelArea()
                    .updateMask(i.select(0).mask())
                    .reduceRegion(ee.Reducer.sum(), g, ee.Number(scale).multiply(10))
                    .values()
                    .get(0)
                )
                return i.set({"maskedFraction": ee.Number(maskedArea).divide(area)})

            images = images.map(f).filter(
                ee.Filter.gt("maskedFraction", options["filterMaskedFraction"])
            )
        else:
            #    get images covering bounds 100%
            images = images.map(
                lambda i: i.set(
                    {
                        "complete": i.select(0)
                        .mask()
                        .reduceRegion(
                            ee.Reducer.allNonZero(), g, ee.Number(scale).multiply(10)
                        )
                        .values()
                        .get(0)
                    }
                )
            ).filter(ee.Filter.eq("complete", 1))

    # exclude night images
    images = images.filter(ee.Filter.gt("SUN_ELEVATION", 0))

    return images


# /***
#  * Sentinel-2 produces multiple images, resulting sometimes 4x more images than the actual size.
#  * This is bad for any statistical analysis.
#  *
#  * This function mosaics images by time.
#  */
def mosaicByTime(images):
    TIME_FIELD = "system:time_start"

    distinct = images.distinct([TIME_FIELD])

    filter = ee.Filter.equals(**{"leftField": TIME_FIELD, "rightField": TIME_FIELD})
    join = ee.Join.saveAll("matches")
    results = join.apply(distinct, images, filter)

    #    mosaic
    def merge_matches(i):
        matchedImages = ee.ImageCollection.fromImages(i.get("matches"))
        mosaic = (
            matchedImages.sort("system:index")
            .mosaic()
            .set({"system:footprint": matchedImages.geometry()})
        )

        return mosaic.copyProperties(i).set(TIME_FIELD, i.get(TIME_FIELD))

    results = results.map(merge_matches)

    return ee.ImageCollection(results)


def mosaic_by_day(images):
    """
    Mosaic the image collection based on the day of the image.
    Args:
        images: input ImageCollection
    Returns:
        ImageCollection with merged images by date.
    """

    images = ee.ImageCollection(images)

    def merge_daily_images(i):
        matches = ee.ImageCollection.fromImages(i.get("images"))
        return (
            matches.sort("system:index")
            .mosaic()
            .copyProperties(i, exclude=["system:time_start"])
            .set(
                {
                    "system:time_start": ee.Date(i.get("date")).millis(),
                    "system:footprint": matches.geometry().dissolve(10),
                }
            )
        )

    def set_image_properties(i):
        d = i.toDictionary()
        return i.set(
            {
                "date": i.date().format("yyyy-MM-dd"),
                "SENSING_ORBIT_NUMBER": ee.Number(d.get("SENSING_ORBIT_NUMBER", -1)),
            }
        )

    images = images.map(set_image_properties)
    distinct = images.distinct(["date"])
    return ee.ImageCollection(
        ee.ImageCollection(
            ee.Join.saveAll("images").apply(
                primary=distinct,
                secondary=images,
                condition=ee.Filter.And(
                    ee.Filter.equals(leftField="date", rightField="date"),
                    ee.Filter.equals(
                        leftField="SPACECRAFT_ID", rightField="SPACECRAFT_ID"
                    ),
                    ee.Filter.equals(
                        leftField="SENSING_ORBIT_NUMBER",
                        rightField="SENSING_ORBIT_NUMBER",
                    ),  # unique for S2
                ),
            )
        ).map(merge_daily_images)
    )


def addQualityScore(images, g, options):
    scorePercentile = 75
    scale = 500
    mask = None
    qualityBand = "green"

    if options:
        if "percentile" in options:
            scorePercentile = options["percentile"]
        if "scale" in options:
            scale = options["scale"]
        if "mask" in options:
            mask = options["mask"]
        if "qualityBand" in options:
            qualityBand = options["qualityBand"]

    def add_quality_score(i):
        score = ee.Image(i).select(
            qualityBand
        )  # //.where(i.select('green').gt(0.5), 0.5)

        if mask:
            score = score.updateMask(mask)

        score = (
            score.reduceRegion(ee.Reducer.percentile([scorePercentile]), g, scale)
            .values()
            .get(0)
        )

        # // var score = i.select('green').add(i.select('blue'))
        # //    .reduceRegion(ee.Reducer.percentile([scorePercentile]), g, scale).values().get(0)

        # // var cloudScore = computeCloudScore(i)
        # // var score = cloudScore.gt(cloudThreshold)
        # //     .reduceRegion(ee.Reducer.sum(), g, scale).values().get(0)

        return i.set({"quality_score": score})

    return images.map(add_quality_score)


def getMostlyCleanImages(images, g, options=None):
    g = ee.Geometry(g)

    scale = 500

    if options and "scale" in options:
        scale = options["scale"]

    p = 85

    if options and "percentile" in options:
        p = options["percentile"]

    # // http://www.earthenv.org/cloud
    modis_clouds = ee.Image("users/gena/MODCF_meanannual")

    # If cloud frequency already known for waterbody
    if options and options.get("cloud_frequency"):
        cloud_frequency = ee.Number(options["cloud_frequency"])
    else:
        cloud_frequency = (
            modis_clouds.divide(10000)
            .reduceRegion(
                ee.Reducer.percentile([p]),
                g.buffer(10000, ee.Number(scale).multiply(10)),
                ee.Number(scale).multiply(10),
            )
            .values()
            .get(0)
        )

    # print('Cloud frequency (over AOI):', cloud_frequency.getInfo())

    # decrease cloudFrequency, include some more partially-cloudy images then clip based on a quality metric
    # also assume inter-annual variability of the cloud cover
    cloud_frequency = ee.Number(cloud_frequency).subtract(0.15).max(0.0)

    if options and options.get("cloudFrequencyThresholdDelta"):
        cloud_frequency = cloud_frequency.add(options["cloudFrequencyThresholdDelta"])

    images = images.filterBounds(g)

    # size = images.size()  # not being used?

    images = addQualityScore(images, g, options).filter(
        ee.Filter.gt("quality_score", 0)
    )  # sometimes null?!

    # clip collection based on quality score and cloud frequency
    # If quality score to filter clouds is already known for waterbody
    if options and options.get("quality_score_cloud_threshold"):
        images = images.filter(
            ee.Filter.lte("quality_score", options["quality_score_cloud_threshold"])
        )
    # Else calculate based on available images
    else:
        images = images.sort("quality_score").limit(
            images.size().multiply(ee.Number(1).subtract(cloud_frequency)).toInt()
        )

    # # remove too dark images
    # images = images.sort('quality_score', false)
    #     .limit(images.size().multiply(0.99).toInt())

    # print('size, filtered: ', images.size())

    return images
    # .set({scoreMax: scoreMax})


def addCdfQualityScore(
    images,
    opt_thresholdMin=75,
    opt_thresholdMax=95,
    opt_includeNeighborhood=False,
    opt_neighborhoodOptions={"erosion": 5, "dilation": 50, "weight": 50},
    opt_qualityBand="green",
):
    """
    Compute one or more high percentile for every pixel and add quality score if pixel value is hither/lower than the threshold %
    """

    images = images.map(lambda i: i.addBands(i.select(opt_qualityBand).rename("q")))

    # P(bad | I < min) = 0, P(bad | I >= max) = 1
    pBad = (
        images.select("q")
        .reduce(ee.Reducer.percentile([opt_thresholdMin, opt_thresholdMax]))
        .rename(["qmin", "qmax"])
    )

    pBadRange = pBad.select("qmax").subtract(pBad.select("qmin"))

    def remove_bad_pixels(i):
        # probability of bad due to high reflactance values
        badLinear = (
            i.select("q")
            .max(pBad.select("qmin"))
            .min(pBad.select("qmax"))
            .subtract(pBad.select("qmin"))
            .divide(pBadRange)
            .clamp(0, 1)
        )

        # too low values
        badLow = i.select("q").lt(0.0001)

        badWeight = badLinear.multiply(badLow.Not())

        if opt_includeNeighborhood:
            radius = opt_neighborhoodOptions

            # opening
            bad = badWeight.gt(0.5)

            if radius.get("erosion"):
                bad = utils.focalMin(bad, radius["erosion"])

            bad = utils.focalMax(bad, radius["dilation"])

            bad = utils.focalMaxWeight(bad, radius["weight"])

            badWeight = badWeight.max(bad)

        # smoothen scene boundaries
        # badWeight = badWeight
        #  .multiply(utils.focalMin(i.select('blue').mask(), 10).convolve(ee.Kernel.gaussian(5, 3)))

        # compute bad pixel probability
        weight = ee.Image(1).float().subtract(badWeight).rename("weight")

        return i.addBands(weight).addBands(pBad)

    # bad pixel removal (costly)
    images = images.map(remove_bad_pixels)

    return images


def otsu(histogram):
    """
    The script computes surface water mask using Canny Edge detector and Otsu thresholding
    See the following paper for details: http://www.mdpi.com/2072-4292/8/5/386

    Author: Gennadii Donchyts (gennadiy.donchyts@gmail.com)
    Contributors: Nicholas Clinton (nclinton@google.com) - re-implemented otsu() using ee.Array

    Example:
        thresholding = require('users/gena/packages:thresholding')

        th = thresholding.computeThresholdUsingOtsu(image, scale, bounds, cannyThreshold, cannySigma, minValue, ...)

    Returns:
        the DN that maximizes interclass variance in B5 (in the region).
    """
    histogram = ee.Dictionary(histogram)

    counts = ee.Array(histogram.get("histogram"))
    means = ee.Array(histogram.get("bucketMeans"))
    size = means.length().get([0])
    total = counts.reduce(ee.Reducer.sum(), [0]).get([0])
    sum = means.multiply(counts).reduce(ee.Reducer.sum(), [0]).get([0])
    mean = sum.divide(total)

    indices = ee.List.sequence(1, size)

    #    Compute between sum of squares, where each mean partitions the data.
    def f(i):
        aCounts = counts.slice(0, 0, i)
        aCount = aCounts.reduce(ee.Reducer.sum(), [0]).get([0])
        aMeans = means.slice(0, 0, i)
        aMean = (
            aMeans.multiply(aCounts)
            .reduce(ee.Reducer.sum(), [0])
            .get([0])
            .divide(aCount)
        )
        bCount = total.subtract(aCount)
        bMean = sum.subtract(aCount.multiply(aMean)).divide(bCount)

        return aCount.multiply(aMean.subtract(mean).pow(2)).add(
            bCount.multiply(bMean.subtract(mean).pow(2))
        )

    bss = indices.map(f)

    #    Return the mean value corresponding to the maximum BSS.
    return means.sort(bss).get([-1])



"""eepackages - A set of utilities built on top of Google Earth Engine (migrated from JavaScript)"""

__version__ = "0.1.0"
__author__ = "Gennadii Donchyts <gennadiy.donchyts@gmail.com>"
__all__ = []



import ee
from eepackages import assets
from eepackages import utils


def computeSurfaceWaterArea(
    waterbody,
    start_filter,
    start,
    stop,
    scale,
    waterOccurrence,
    opt_missions=None,
    quality_score_attributes=None,
):
    """
    compute the surface area of given waterbody between start and stop, using images between
    start_filter and stop to obtain a representative set of images to filter cloudy images.

    args:
      waterbody: waterbody feature
      start_filter: start date of images using to filter cloudy images
      start: start date of analysis
      stop: stop date of analysis
      scale: scale of analysis
      waterOccurrence: feature collection of reference water occurrence data.
      opt_missions: mission codes to use for the input imagery. Leave empty for a recommended set of
        missions.
      quality_score_attribute: if the quality_score is precomputed, a ranked list of attributes of
        the waterbody to use as quality_score. Otherwise None.
    returns:
      image collection containing water layers.
    """

    geom = ee.Feature(waterbody).geometry()

    missions = ["L4", "L5", "L7", "L8", "S2"]
    props = waterbody.getInfo()["properties"]
    quality_score_threshold = None
    if quality_score_attributes:
        for prop in quality_score_attributes:
            att = props.get(prop)
            if att:
                quality_score_threshold = att
                break

    if opt_missions:
        missions = opt_missions

    images = assets.getImages(
        geom,
        {
            "resample": True,
            "filter": ee.Filter.date(start_filter, stop),
            "missions": missions,
            "scale": scale * 10,
        },
    )

    # TODO: mosaic by day
    # images = assets.mosaic_by_day(images)

    options = {
        # 'cloudFrequencyThresholdDelta': -0.15
        "scale": scale * 5,
        "cloud_frequency": props.get("cloud_frequency"),
        "quality_score_cloud_threshold": quality_score_threshold,
    }

    g = geom.buffer(300, scale)

    images = assets.getMostlyCleanImages(images, g, options).sort("system:time_start")
    images = images.filterDate(start, stop)

    # all images, but with quality score
    # images = (assets.addQualityScore(images, g, options)
    #   .filter(ee.Filter.gt('quality_score', 0))) # sometimes null?!

    # print('Image count (clean): ', images.size())

    water = images.map(
        lambda i: computeSurfaceWaterArea_SingleImage(
            i, waterbody, scale, waterOccurrence
        )
    )

    water = water.filter(ee.Filter.neq("area", 0))

    return water


def computeSurfaceWaterArea_SingleImage(i, waterbody, scale, waterOccurrence):
    geom = ee.Feature(waterbody).geometry()

    fillPercentile = 50  # // we don't trust our prior

    ndwiBands = ["green", "swir"]
    # var ndwiBands = ['green', 'nir']

    waterMaxImage = ee.Image().float().paint(waterbody.buffer(150), 1)

    maxArea = waterbody.area(scale)

    t = i.get("system:time_start")

    i = i.updateMask(waterMaxImage).updateMask(
        i.select("swir").min(i.select("nir")).gt(0.001)
    )

    ndwi = i.normalizedDifference(ndwiBands)

    # var water = ndwi.gt(0)

    th = utils.computeThresholdUsingOtsu(ndwi, scale, geom, 0.5, 0.7, -0.2)
    water = ndwi.gt(th)

    area = (
        ee.Image.pixelArea()
        .mask(water)
        .reduceRegion(**{"reducer": ee.Reducer.sum(), "geometry": geom, "scale": scale})
        .get("area")
    )

    # # fill missing, estimate water probability as the lowest percentile.
    # var waterEdge = ee.Algorithms.CannyEdgeDetector(water, 0.1, 0)
    waterEdge = ee.Algorithms.CannyEdgeDetector(ndwi, 0.5, 0.7)

    # image mask
    imageMask = ndwi.mask()

    # var imageMask = i.select(ndwiBands).reduce(ee.Reducer.allNonZero())

    # imageMask = utils.focalMin(imageMask, ee.Number(scale) * 1.5)
    imageMask = imageMask.focal_min(ee.Number(scale).multiply(1.5), "square", "meters")
    # .multiply(waterMaxImage)

    # TODO: exclude non-water/missing boundsry
    waterEdge = waterEdge.updateMask(imageMask)

    # # clip by clouds
    # var bad = ee.Image(1).float().subtract(i.select('weight'))
    # bad = utils.focalMax(bad, 90).not()
    # waterEdge = waterEdge.updateMask(bad)

    # get water probability around edges
    # P(D|W) = P(D|W) * P(W) / P(D) ~=  P(D|W) * P(W)
    p = (
        waterOccurrence.mask(waterEdge)
        .reduceRegion(
            **{
                "reducer": ee.Reducer.percentile([fillPercentile]),
                "geometry": geom,
                "scale": scale,
            }
        )
        .values()
        .get(0)
    )

    # TODO: exclude edges belonging to cloud/water or cloud/land

    # TODO: use multiple percentiles (confidence margin)

    p = ee.Algorithms.If(ee.Algorithms.IsEqual(p, None), 101, p)

    waterFill = waterOccurrence.gt(ee.Image.constant(p)).updateMask(
        water.unmask(0, False).Not()
    )

    # exclude false-positive, where we're sure in a non-water
    nonWater = ndwi.lt(-0.15).unmask(0, False)
    waterFill = waterFill.updateMask(nonWater.Not())

    fill = (
        ee.Image.pixelArea()
        .mask(waterFill)
        .reduceRegion(**{"reducer": ee.Reducer.sum(), "geometry": geom, "scale": scale})
        .get("area")
    )

    area_filled = ee.Number(area).add(fill)

    filled_fraction = ee.Number(fill).divide(area_filled)

    return (
        i.addBands(waterFill.rename("water_fill"))
        .addBands(waterEdge.rename("water_edge"))
        .addBands(ndwi.rename("ndwi"))
        .addBands(water.rename("water"))
        .set(
            {
                "p": p,
                "area": area,
                "area_filled": area_filled,
                "filled_fraction": filled_fraction,
                "system:time_start": t,
                "ndwi_threshold": th,
                "waterOccurrenceExpected": waterOccurrence.mask(waterEdge),
            }
        )
    )


def computeSurfaceWaterAreaJRC(waterbody, start, stop, scale):
    geom = ee.Feature(waterbody).geometry()

    jrcMonthly = ee.ImageCollection("JRC/GSW1_0/MonthlyHistory")

    def compute_area(i):
        area = (
            ee.Image.pixelArea()
            .mask(i.eq(2))
            .reduceRegion(
                {"reducer": ee.Reducer.sum(), "geometry": geom, "scale": scale}
            )
        )

        return i.set({"area": area.get("area")})

    water = jrcMonthly.filterDate(start, stop).map(compute_area)

    return water


def extrapolate_JRC_wo(waterbody, scale, max_trusted_occurrence=0.97):
    """
    extrapolate the JRC dataset based on a maximum value that is trusted.
    This is needed when a waterbody experiences an unprecedented drought, in which gap filling
    fails when using the JRC dataset as-is.
    """
    water_occurrence = (
        ee.Image("JRC/GSW1_4/GlobalSurfaceWater")
        .select("occurrence")
        .mask(1)  # Mask important to fix JRC inconsistent masking.
        .resample("bicubic")
        .divide(100)
    )

    # we don't trust water occurrence dataset when water level is below this (values > th)
    min_water = water_occurrence.gt(max_trusted_occurrence)

    # Calculate distance from water Occurence max
    dist = (
        min_water.Not()
        .fastDistanceTransform(150)
        .reproject(ee.Projection("EPSG:4326").atScale(100))
        .resample("bicubic")
        .sqrt()
        .multiply(100)
    )

    # calculate max distance for scaling to 1
    max_distance = dist.reduceRegion(
        reducer=ee.Reducer.max(),
        geometry=waterbody.geometry(),
        scale=scale,
        bestEffort=True,
    ).get("distance")

    # scale distance values from min_trusted_occurrence to 1
    extrap_scale = ee.Number(1).subtract(max_trusted_occurrence).divide(max_distance)
    water_occurrence_extrapolated = dist.multiply(extrap_scale).add(
        max_trusted_occurrence
    )

    return water_occurrence.where(
        water_occurrence.gt(max_trusted_occurrence), water_occurrence_extrapolated
    )



import ee
import math


def createTransectAtCentroid(line, length, crs, maxError):
    """
    Creates transect in the middle of line geometry
    """
    if not crs:
        crs = "EPSG:3857"

    if not maxError:
        maxError = 1

    line = ee.Geometry(line).transform(crs, maxError)
    origin = line.centroid(maxError, crs)

    length = ee.Number(length)

    # compute angle from two points
    coords = line.coordinates()
    pt0 = coords.slice(0, 1)
    pt1 = coords.slice(-1)
    delta = ee.Array(pt1).subtract(pt0)
    dx = delta.get([0, 0])
    dy = delta.get([0, 1])
    angle = dx.atan2(dy).add(math.pi / 2)

    ptOrigin = ee.Array([origin.coordinates()]).transpose()

    # get coordinates as a list
    proj1 = origin.projection().translate(length.multiply(-0.5), 0)
    pt1 = ee.Array([origin.transform(proj1).coordinates()]).transpose()

    # translate
    proj2 = origin.projection().translate(length.multiply(0.5), 0)
    pt2 = ee.Array([origin.transform(proj2).coordinates()]).transpose()

    # define rotation matrix
    cosa = angle.cos()
    sina = angle.sin()
    M = ee.Array([[cosa, sina.multiply(-1)], [sina, cosa]])

    # rotate
    pt1 = M.matrixMultiply(pt1.subtract(ptOrigin)).add(ptOrigin)
    pt2 = M.matrixMultiply(pt2.subtract(ptOrigin)).add(ptOrigin)

    # get points
    pt1 = pt1.transpose().project([1]).toList()
    pt2 = pt2.transpose().project([1]).toList()

    # construct line
    line = ee.Algorithms.GeometryConstructors.LineString([pt1, pt2], ee.Projection(crs))

    return line


def createVector(origin, angle, length):
    # get coordinates as a list
    pt1 = ee.Array([ee.Feature(origin).geometry().coordinates()]).transpose()

    # translate
    proj = origin.projection().translate(length, 0)
    pt2 = ee.Array([origin.transform(proj).coordinates()]).transpose()

    # define rotation matrix
    angle = ee.Number(angle).multiply(math.pi).divide(180)
    cosa = angle.cos()
    sina = angle.sin()
    M = ee.Array([[cosa, sina.multiply(-1)], [sina, cosa]])

    # rotate
    pt2 = M.matrixMultiply(pt2.subtract(pt1)).add(pt1)

    # get end point
    pt2 = pt2.transpose().project([1]).toList()

    # construct line
    line = ee.Algorithms.GeometryConstructors.LineString(
        [origin, pt2], origin.projection()
    )

    return line


def angle(pt0, pt1):
    pt0 = ee.List(pt0)
    pt1 = ee.List(pt1)

    x0 = ee.Number(pt0.get(0))
    x1 = ee.Number(pt1.get(0))
    y0 = ee.Number(pt0.get(1))
    y1 = ee.Number(pt1.get(1))

    dy = y1.subtract(y0)
    dx = x1.subtract(x0)

    return dy.atan2(dx)



import io
import os
import re

from setuptools import find_packages
from setuptools import setup

__version__ = "0.25.0"


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type("")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="eepackages",
    version=__version__,
    url="https://github.com/gee-community/ee-packages-py",
    license="MIT",
    author="Gennadii Donchyts",
    author_email="gennadiy.donchyts@gmail.com",
    description="A set of utilities built on top of Google Earth Engine (migrated from JavaScript)",
    long_description=read("README.md"),
    packages=find_packages(exclude=("tests",)),
    install_requires=["earthengine-api>=0.1.284", "pathos>=0.2.8", "retry>=0.9.2"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



from datetime import timedelta
from itertools import repeat
import math
from pathlib import Path
from time import sleep, time
from typing import Any, Callable, Dict, Optional

import ee
from pathos.multiprocessing import ProcessPool

from eepackages.multiprocessing.download_image import download_image

# /***
#  * The script computes surface water mask using Canny Edge detector and Otsu thresholding
#  * See the following paper for details: http://www.mdpi.com/2072-4292/8/5/386
#  *
#  * Author: Gennadii Donchyts (gennadiy.donchyts@gmail.com)
#  * Contributors: Nicholas Clinton (nclinton@google.com) - re-implemented otsu() using ee.Array
#  *
#  * Usage:
#  *
#  * var thresholding = require('users/gena/packages:thresholding')
#  *
#  * var th = thresholding.computeThresholdUsingOtsu(image, scale, bounds, cannyThreshold, cannySigma, minValue, ...)
#  *
#  */

#  /***
#  * Return the DN that maximizes interclass variance in B5 (in the region).
#  */


def otsu(histogram):
    histogram = ee.Dictionary(histogram)

    counts = ee.Array(histogram.get("histogram"))
    means = ee.Array(histogram.get("bucketMeans"))
    size = means.length().get([0])
    total = counts.reduce(ee.Reducer.sum(), [0]).get([0])
    sum = means.multiply(counts).reduce(ee.Reducer.sum(), [0]).get([0])
    mean = sum.divide(total)

    indices = ee.List.sequence(1, size)

    #  Compute between sum of squares, where each mean partitions the data.
    def f(i):
        aCounts = counts.slice(0, 0, i)
        aCount = aCounts.reduce(ee.Reducer.sum(), [0]).get([0])
        aMeans = means.slice(0, 0, i)
        aMean = (
            aMeans.multiply(aCounts)
            .reduce(ee.Reducer.sum(), [0])
            .get([0])
            .divide(aCount)
        )
        bCount = total.subtract(aCount)
        bMean = sum.subtract(aCount.multiply(aMean)).divide(bCount)

        return aCount.multiply(aMean.subtract(mean).pow(2)).add(
            bCount.multiply(bMean.subtract(mean).pow(2))
        )

    bss = indices.map(f)

    #  Return the mean value corresponding to the maximum BSS.
    return means.sort(bss).get([-1])


# /***
#  * Compute a threshold using Otsu method (bimodal)
#  */


def computeThresholdUsingOtsu(
    image,
    scale,
    bounds,
    cannyThreshold,
    cannySigma,
    minValue,
    debug=False,
    minEdgeLength=None,
    minEdgeGradient=None,
    minEdgeValue=None,
):
    #  clip image edges
    mask = (
        image.mask()
        .gt(0)
        .clip(bounds)
        .focal_min(ee.Number(scale).multiply(3), "circle", "meters")
    )

    #  detect sharp changes
    edge = ee.Algorithms.CannyEdgeDetector(image, cannyThreshold, cannySigma)
    edge = edge.multiply(mask)

    if minEdgeLength:
        connected = edge.mask(edge).lt(cannyThreshold).connectedPixelCount(200, True)

        edgeLong = connected.gte(minEdgeLength)

        # if debug:
        #  print('Edge length: ', ui.Chart.image.histogram(connected, bounds, scale, buckets))
        #  Map.addLayer(edge.mask(edge), {palette:['ff0000']}, 'edges (short)', false);

        edge = edgeLong

    #  buffer around NDWI edges
    edgeBuffer = edge.focal_max(ee.Number(scale), "square", "meters")

    if minEdgeValue:
        edgeMin = image.reduceNeighborhood(
            ee.Reducer.min(), ee.Kernel.circle(ee.Number(scale), "meters")
        )

        edgeBuffer = edgeBuffer.updateMask(edgeMin.gt(minEdgeValue))

        # if debug:
        #  Map.addLayer(edge.updateMask(edgeBuffer), {palette:['ff0000']}, 'edge min', false);

    if minEdgeGradient:
        edgeGradient = (
            image.gradient()
            .abs()
            .reduce(ee.Reducer.max())
            .updateMask(edgeBuffer.mask())
        )

        edgeGradientTh = ee.Number(
            edgeGradient.reduceRegion(
                ee.Reducer.percentile([minEdgeGradient]), bounds, scale
            )
            .values()
            .get(0)
        )

        # if debug:
        #  print('Edge gradient threshold: ', edgeGradientTh)

        #  Map.addLayer(edgeGradient.mask(edgeGradient), {palette:['ff0000']}, 'edge gradient', false);

        #  print('Edge gradient: ', ui.Chart.image.histogram(edgeGradient, bounds, scale, buckets))

        edgeBuffer = edgeBuffer.updateMask(edgeGradient.gt(edgeGradientTh))

    edge = edge.updateMask(edgeBuffer)
    edgeBuffer = edge.focal_max(ee.Number(scale).multiply(1), "square", "meters")
    imageEdge = image.mask(edgeBuffer)

    # if debug:
    #  Map.addLayer(imageEdge, {palette:['222200', 'ffff00']}, 'image edge buffer', false)

    #  compute threshold using Otsu thresholding
    buckets = 100
    hist = ee.Dictionary(
        ee.Dictionary(
            imageEdge.reduceRegion(ee.Reducer.histogram(buckets), bounds, scale)
        )
        .values()
        .get(0)
    )

    threshold = ee.Algorithms.If(hist.contains("bucketMeans"), otsu(hist), minValue)
    threshold = ee.Number(threshold)

    if debug:
        # // experimental
        # // var jrc = ee.Image('JRC/GSW1_0/GlobalSurfaceWater').select('occurrence')
        # // var jrcTh = ee.Number(ee.Dictionary(jrc.updateMask(edge).reduceRegion(ee.Reducer.mode(), bounds, scale)).values().get(0))
        # // var water = jrc.gt(jrcTh)
        # // Map.addLayer(jrc, {palette: ['000000', 'ffff00']}, 'JRC')
        # // print('JRC occurrence (edge)', ui.Chart.image.histogram(jrc.updateMask(edge), bounds, scale, buckets))

        # Map.addLayer(edge.mask(edge), {palette:['ff0000']}, 'edges', true);

        print("Threshold: ", threshold)

        # print('Image values:', ui.Chart.image.histogram(image, bounds, scale, buckets));
        # print('Image values (edge): ', ui.Chart.image.histogram(imageEdge, bounds, scale, buckets));
        # Map.addLayer(mask.mask(mask), {palette:['000000']}, 'image mask', false);

    if minValue is not None:
        return threshold.max(minValue)
    else:
        return threshold


def focalMin(image: ee.Image, radius: float):
    erosion: ee.Image = (
        image.Not().fastDistanceTransform(radius).sqrt().lte(radius).Not()
    )
    return erosion


def focalMax(image: ee.Image, radius: float):
    dilation: ee.Image = image.fastDistanceTransform(radius).sqrt().lte(radius)
    return dilation


def focalMaxWeight(image: ee.Image, radius: float):
    distance: ee.Image = image.fastDistanceTransform(radius).sqrt()
    dilation: ee.Image = distance.where(distance.gte(radius), radius)
    dilation = ee.Image(radius).subtract(dilation).divide(radius)
    return dilation


def _batch_download_ic(
    ic: ee.ImageCollection,
    img_download_method: Callable,
    name_prefix: str,
    out_dir: Optional[Path],
    pool_size: int,
    download_kwargs: Optional[Dict[str, Any]],
):
    """
    does the actual work batch downloading images in an ee.ImageCollection using the python
    multiprocessing module. Takes the img download method as a callable to implement different
    ee.Image methods.
    """

    out_dir.mkdir(exist_ok=True, parents=True)
    log_dir: Path = out_dir / "logging"
    log_dir.mkdir(exist_ok=True, parents=True)

    num_images: int = ic.size().getInfo()
    image_list: ee.List = ic.toList(num_images)

    serialized_il: str = image_list.serialize()

    pool: ProcessPool = ProcessPool(nodes=pool_size)
    result = pool.amap(
        download_image,
        repeat(ee),
        range(num_images),
        repeat(img_download_method),
        repeat(serialized_il),
        repeat(name_prefix),
        repeat(out_dir),
        repeat(download_kwargs),
    )

    start_time: time = time()
    elapsed: timedelta = timedelta(0)
    result.get()  # needed to trigger execution?

    while not result.ready() and elapsed.total_seconds() < 3600:  # 1h timeout
        sleep(5)
        elapsed = timedelta(seconds=time() - start_time)

    pool.close()
    pool.join()
    pool.clear()


def download_image_collection(
    ic: ee.ImageCollection,
    name_prefix: str = "ic_",
    out_dir: Optional[Path] = None,
    pool_size: int = 25,
    download_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Download images in image collection. Only works for images in the collection that are < 32M
    and grid dimension < 10000, documented at
    https://developers.google.com/earth-engine/apidocs/ee-image-getdownloadurl.

    args:
        ic (ee.ImageCollection): ImageCollection to download.
        name_prefix (str): prefix for the filename of the downloaded objects.
        out_dir (Optional(Path)): pathlib object referring to output dir.
        pool_size (int): multiprocessing pool size.
        download_kwargs (Optional(Dict(str, Any))): keyword arguments used in
            [getDownloadUrl](https://developers.google.com/earth-engine/apidocs/ee-image-getdownloadurl).
    """
    _batch_download_ic(
        ic, ee.Image.getDownloadURL, name_prefix, out_dir, pool_size, download_kwargs
    )


def download_image_collection_thumb(
    ic: ee.ImageCollection,
    name_prefix: str = "ic_",
    out_dir: Optional[Path] = None,
    pool_size: int = 25,
    download_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Download thumb images in and image collection. Only works for images in the collection that are < 32M
    and grid dimension < 10000, documented at
    https://developers.google.com/earth-engine/apidocs/ee-image-getthumburl.

    args:
        ic (ee.ImageCollection): ImageCollection to download.
        name_prefix (str): prefix for the filename of the downloaded objects.
        out_dir (Optional(Path)): pathlib object referring to output dir.
        pool_size (int): multiprocessing pool size.
        download_kwargs (Optional(Dict(str, Any))): keyword arguments used in
            [getDownloadUrl](https://developers.google.com/earth-engine/apidocs/ee-image-getthumburl).
    """
    _batch_download_ic(
        ic, ee.Image.getThumbURL, name_prefix, out_dir, pool_size, download_kwargs
    )


def radians(img):
    """Converts image from degrees to radians"""
    return img.toFloat().multiply(math.pi).divide(180)


def hillshade(az, ze, slope, aspect):
    """Computes hillshade"""
    azimuth = radians(ee.Image.constant(az))
    zenith = radians(ee.Image.constant(90).subtract(ee.Image.constant(ze)))

    return (
        azimuth.subtract(aspect)
        .cos()
        .multiply(slope.sin())
        .multiply(zenith.sin())
        .add(zenith.cos().multiply(slope.cos()))
    )


def hillshadeRGB(
    image,
    elevation,
    weight=1,
    height_multiplier=5,
    azimuth=0,
    zenith=45,
    contrast=0,
    brightness=0,
    saturation=1,
    castShadows=False,
    customTerrain=False,
):
    """Styles RGB image using hillshading, mixes RGB and hillshade using HSV<->RGB transform"""

    hsv = image.visualize().unitScale(0, 255).rgbToHsv()

    z = elevation.multiply(ee.Image.constant(height_multiplier))

    terrain = ee.Algorithms.Terrain(z)
    slope = radians(terrain.select(["slope"])).resample("bicubic")
    aspect = radians(terrain.select(["aspect"])).resample("bicubic")

    if customTerrain:
        raise NotImplementedError("customTerrain argument is not implemented yet")

    hs = hillshade(azimuth, zenith, slope, aspect).resample("bicubic")

    if castShadows:
        hysteresis = True
        neighborhoodSize = 256

        hillShadow = ee.Algorithms.HillShadow(
            elevation,
            azimuth,
            ee.Number(90).subtract(zenith),
            neighborhoodSize,
            hysteresis,
        ).float()

        hillShadow = ee.Image(1).float().subtract(hillShadow)

        # opening
        # hillShadow = hillShadow.multiply(hillShadow.focal_min(3).focal_max(6))

        # cleaning
        hillShadow = hillShadow.focal_mode(3)

        # smoothing
        hillShadow = hillShadow.convolve(ee.Kernel.gaussian(5, 3))

        # transparent
        hillShadow = hillShadow.multiply(0.7)

        hs = hs.subtract(hillShadow).rename("shadow")

    intensity = hs.multiply(ee.Image.constant(weight)).add(
        hsv.select("value").multiply(ee.Image.constant(1).subtract(weight))
    )

    sat = hsv.select("saturation").multiply(saturation)

    hue = hsv.select("hue")

    result = (
        ee.Image.cat(hue, sat, intensity)
        .hsvToRgb()
        .multiply(ee.Image.constant(1).float().add(contrast))
        .add(ee.Image.constant(brightness).float())
    )

    if customTerrain:
        mask = elevation.mask().focal_min(2)

        result = result.updateMask(mask)

    return result


def getIsolines(image, levels=None):
    """Adds isolines to an ee.Image"""

    def addIso(image, level):
        crossing = image.subtract(level).focal_median(3).zeroCrossing()
        exact = image.eq(level)
        return ee.Image(level).float().mask(crossing.Or(exact)).set({"level": level})

    if not levels:
        levels = ee.List.sequence(0, 1, 0.1)

    levels = ee.List(levels)

    isoImages = ee.ImageCollection(levels.map(lambda l: addIso(image, ee.Number(l))))

    return isoImages
